@@ -1,15 +1,17 @@
-"""Tests for aiohttp/server.py"""
+# Tests for aiohttp/server.py
 
 import asyncio
+import platform
 import socket
 from functools import partial
-from html import escape
 from unittest import mock
 
 import pytest
 
 from aiohttp import helpers, http, streams, web
 
+IS_MACOS = platform.system() == "Darwin"
+
 
 @pytest.fixture
 def make_srv(loop, manager):
@@ -17,7 +19,7 @@ def make_srv(loop, manager):
 
     def maker(*, cls=web.RequestHandler, **kwargs):
         nonlocal srv
-        m = kwargs.pop('manager', manager)
+        m = kwargs.pop("manager", manager)
         srv = cls(m, loop=loop, access_log=None, **kwargs)
         return srv
 
@@ -32,6 +34,7 @@ def make_srv(loop, manager):
 def manager(request_handler, loop):
     async def maker():
         return web.Server(request_handler)
+
     return loop.run_until_complete(maker())
 
 
@@ -40,9 +43,10 @@ def srv(make_srv, transport):
     srv = make_srv()
     srv.connection_made(transport)
     transport.close.side_effect = partial(srv.connection_lost, None)
-    srv._drain_helper = mock.Mock()
-    srv._drain_helper.side_effect = helpers.noop
-    return srv
+    with mock.patch.object(
+        web.RequestHandler, "_drain_helper", side_effect=helpers.noop
+    ):
+        yield srv
 
 
 @pytest.fixture
@@ -52,7 +56,6 @@ def buf():
 
 @pytest.fixture
 def request_handler():
-
     async def handler(request):
         return web.Response()
 
@@ -64,13 +67,13 @@ def request_handler():
 @pytest.fixture
 def handle_with_error():
     def wrapper(exc=ValueError):
-
         async def handle(request):
             raise exc
 
         h = mock.Mock()
         h.side_effect = handle
         return h
+
     return wrapper
 
 
@@ -92,14 +95,6 @@ def transport(buf):
     return transport
 
 
-@pytest.fixture
-def ceil(mocker):
-    def ceil(val):
-        return val
-
-    mocker.patch('aiohttp.helpers.ceil').side_effect = ceil
-
-
 async def test_shutdown(srv, transport) -> None:
     loop = asyncio.get_event_loop()
     assert transport is srv.transport
@@ -114,7 +109,7 @@ async def test_shutdown(srv, transport) -> None:
     await srv.shutdown()
     t1 = loop.time()
 
-    assert t1 - t0 < 0.05, t1-t0
+    assert t1 - t0 < 0.05, t1 - t0
 
     assert transport.close.called
     assert srv.transport is None
@@ -148,9 +143,8 @@ async def test_shutdown_wait_error_handler(srv, transport) -> None:
 
 async def test_close_after_response(srv, transport) -> None:
     srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
+        b"GET / HTTP/1.0\r\n" b"Host: example.com\r\n" b"Content-Length: 0\r\n\r\n"
+    )
     h = srv._task_handler
 
     await asyncio.sleep(0.1)
@@ -175,8 +169,7 @@ def test_connection_made_with_tcp_keepaplive(make_srv, transport) -> None:
     sock = mock.Mock()
     transport.get_extra_info.return_value = sock
     srv.connection_made(transport)
-    sock.setsockopt.assert_called_with(socket.SOL_SOCKET,
-                                       socket.SO_KEEPALIVE, 1)
+    sock.setsockopt.assert_called_with(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
 
 
 def test_connection_made_without_tcp_keepaplive(make_srv) -> None:
@@ -198,9 +191,8 @@ def test_eof_received(make_srv) -> None:
 
 async def test_connection_lost(srv) -> None:
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
+        b"GET / HTTP/1.1\r\n" b"Host: example.com\r\n" b"Content-Length: 0\r\n\r\n"
+    )
     srv._keepalive = True
 
     handle = srv._task_handler
@@ -234,84 +226,36 @@ def test_srv_keep_alive_disable(srv) -> None:
 
 
 async def test_simple(srv, buf) -> None:
-    srv.data_received(
-        b'GET / HTTP/1.1\r\n\r\n')
+    srv.data_received(b"GET / HTTP/1.1\r\n\r\n")
 
     await asyncio.sleep(0.05)
-    assert buf.startswith(b'HTTP/1.1 200 OK\r\n')
+    assert buf.startswith(b"HTTP/1.1 200 OK\r\n")
 
 
 async def test_bad_method(srv, buf) -> None:
-    srv.data_received(
-        b':BAD; / HTTP/1.0\r\n'
-        b'Host: example.com\r\n\r\n')
+    srv.data_received(b":BAD; / HTTP/1.0\r\n" b"Host: example.com\r\n\r\n")
 
     await asyncio.sleep(0)
-    assert buf.startswith(b'HTTP/1.0 400 Bad Request\r\n')
-
-
-async def test_data_received_error(srv, buf) -> None:
-    transport = srv.transport
-    srv._request_parser = mock.Mock()
-    srv._request_parser.feed_data.side_effect = TypeError
-
-    srv.data_received(
-        b'!@#$ / HTTP/1.0\r\n'
-        b'Host: example.com\r\n\r\n')
-
-    await asyncio.sleep(0)
-    assert buf.startswith(b'HTTP/1.0 500 Internal Server Error\r\n')
-    assert transport.close.called
-    assert srv._error_handler is None
+    assert buf.startswith(b"HTTP/1.0 400 Bad Request\r\n")
 
 
 async def test_line_too_long(srv, buf) -> None:
-    srv.data_received(b''.join([b'a' for _ in range(10000)]) + b'\r\n\r\n')
+    srv.data_received(b"".join([b"a" for _ in range(10000)]) + b"\r\n\r\n")
 
     await asyncio.sleep(0)
-    assert buf.startswith(b'HTTP/1.0 400 Bad Request\r\n')
+    assert buf.startswith(b"HTTP/1.0 400 Bad Request\r\n")
 
 
 async def test_invalid_content_length(srv, buf) -> None:
     srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: sdgg\r\n\r\n')
+        b"GET / HTTP/1.0\r\n" b"Host: example.com\r\n" b"Content-Length: sdgg\r\n\r\n"
+    )
     await asyncio.sleep(0)
 
-    assert buf.startswith(b'HTTP/1.0 400 Bad Request\r\n')
+    assert buf.startswith(b"HTTP/1.0 400 Bad Request\r\n")
 
 
-async def test_handle_error__utf(
-    make_srv, buf, transport, request_handler
-):
-    request_handler.side_effect = RuntimeError('что-то пошло не так')
-
-    srv = make_srv(debug=True)
-    srv.connection_made(transport)
-    srv.keep_alive(True)
-    srv.logger = mock.Mock()
-
-    srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
-    await asyncio.sleep(0)
-
-    assert b'HTTP/1.0 500 Internal Server Error' in buf
-    assert b'Content-Type: text/html; charset=utf-8' in buf
-    pattern = escape("RuntimeError: что-то пошло не так")
-    assert pattern.encode('utf-8') in buf
-    assert not srv._keepalive
-
-    srv.logger.exception.assert_called_with(
-        "Error handling request", exc_info=mock.ANY)
-
-
-async def test_unhandled_runtime_error(
-    make_srv, transport, request_handler
-):
-
+async def test_unhandled_runtime_error(make_srv, transport, request_handler):
     async def handle(request):
         resp = web.Response()
         resp.write_eof = mock.Mock()
@@ -325,18 +269,19 @@ async def test_unhandled_runtime_error(
     request_handler.side_effect = handle
 
     srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
+        b"GET / HTTP/1.0\r\n" b"Host: example.com\r\n" b"Content-Length: 0\r\n\r\n"
+    )
 
     await srv._task_handler
     assert request_handler.called
     srv.logger.exception.assert_called_with(
-        "Unhandled runtime exception", exc_info=mock.ANY)
+        "Unhandled runtime exception", exc_info=mock.ANY
+    )
 
 
 async def test_handle_uncompleted(
-        make_srv, transport, handle_with_error, request_handler):
+    make_srv, transport, handle_with_error, request_handler
+):
     closed = False
 
     def close():
@@ -351,19 +296,24 @@ async def test_handle_uncompleted(
     request_handler.side_effect = handle_with_error()
 
     srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 50000\r\n\r\n')
+        b"GET / HTTP/1.0\r\n" b"Host: example.com\r\n" b"Content-Length: 50000\r\n\r\n"
+    )
 
     await srv._task_handler
     assert request_handler.called
     assert closed
-    srv.logger.exception.assert_called_with(
-        "Error handling request", exc_info=mock.ANY)
+    srv.logger.exception.assert_called_with("Error handling request", exc_info=mock.ANY)
 
 
+@pytest.mark.xfail(
+    IS_MACOS,
+    raises=TypeError,
+    reason="Intermittently fails on macOS",
+    strict=False,
+)
 async def test_handle_uncompleted_pipe(
-        make_srv, transport, request_handler, handle_with_error):
+    make_srv, transport, request_handler, handle_with_error
+):
     closed = False
     normal_completed = False
 
@@ -386,28 +336,25 @@ async def test_handle_uncompleted_pipe(
     # normal
     request_handler.side_effect = handle
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
-    await asyncio.sleep(0)
+        b"GET / HTTP/1.1\r\n" b"Host: example.com\r\n" b"Content-Length: 0\r\n\r\n"
+    )
+    await asyncio.sleep(0.01)
 
     # with exception
     request_handler.side_effect = handle_with_error()
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 50000\r\n\r\n')
+        b"GET / HTTP/1.1\r\n" b"Host: example.com\r\n" b"Content-Length: 50000\r\n\r\n"
+    )
 
     assert srv._task_handler
 
-    await asyncio.sleep(0)
+    await asyncio.sleep(0.01)
 
     await srv._task_handler
     assert normal_completed
     assert request_handler.called
     assert closed
-    srv.logger.exception.assert_called_with(
-        "Error handling request", exc_info=mock.ANY)
+    srv.logger.exception.assert_called_with("Error handling request", exc_info=mock.ANY)
 
 
 async def test_lingering(srv, transport) -> None:
@@ -416,24 +363,23 @@ async def test_lingering(srv, transport) -> None:
     async def handle(message, request, writer):
         pass
 
-    srv.handle_request = handle
-    srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 3\r\n\r\n')
+    with mock.patch.object(
+        web.RequestHandler, "handle_request", create=True, new=handle
+    ):
+        srv.data_received(
+            b"GET / HTTP/1.0\r\n" b"Host: example.com\r\n" b"Content-Length: 3\r\n\r\n"
+        )
 
-    await asyncio.sleep(0.05)
-    assert not transport.close.called
+        await asyncio.sleep(0.05)
+        assert not transport.close.called
 
-    srv.data_received(b'123')
+        srv.data_received(b"123")
 
-    await asyncio.sleep(0)
-    transport.close.assert_called_with()
+        await asyncio.sleep(0)
+        transport.close.assert_called_with()
 
 
-async def test_lingering_disabled(make_srv,
-                                  transport, request_handler) -> None:
-
+async def test_lingering_disabled(make_srv, transport, request_handler) -> None:
     async def handle_request(request):
         await asyncio.sleep(0)
 
@@ -445,19 +391,15 @@ async def test_lingering_disabled(make_srv,
     assert not transport.close.called
 
     srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 50\r\n\r\n')
+        b"GET / HTTP/1.0\r\n" b"Host: example.com\r\n" b"Content-Length: 50\r\n\r\n"
+    )
     await asyncio.sleep(0)
     assert not transport.close.called
     await asyncio.sleep(0.05)
     transport.close.assert_called_with()
 
 
-async def test_lingering_timeout(
-    make_srv, transport, ceil, request_handler
-):
-
+async def test_lingering_timeout(make_srv, transport, request_handler):
     async def handle_request(request):
         await asyncio.sleep(0)
 
@@ -469,9 +411,8 @@ async def test_lingering_timeout(
     assert not transport.close.called
 
     srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 50\r\n\r\n')
+        b"GET / HTTP/1.0\r\n" b"Host: example.com\r\n" b"Content-Length: 50\r\n\r\n"
+    )
     await asyncio.sleep(0)
     assert not transport.close.called
 
@@ -479,15 +420,11 @@ async def test_lingering_timeout(
     transport.close.assert_called_with()
 
 
-async def test_handle_payload_access_error(
-    make_srv, transport, request_handler
-):
+async def test_handle_payload_access_error(make_srv, transport, request_handler):
     srv = make_srv(lingering_time=0)
     srv.connection_made(transport)
     srv.data_received(
-        b'POST /test HTTP/1.1\r\n'
-        b'Content-Length: 9\r\n\r\n'
-        b'some data'
+        b"POST /test HTTP/1.1\r\n" b"Content-Length: 9\r\n\r\n" b"some data"
     )
     # start request_handler task
     await asyncio.sleep(0.05)
@@ -505,18 +442,18 @@ async def test_handle_cancel(make_srv, transport) -> None:
     async def handle_request(message, payload, writer):
         await asyncio.sleep(10)
 
-    srv.handle_request = handle_request
-
     async def cancel():
         srv._task_handler.cancel()
 
-    srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Content-Length: 10\r\n'
-        b'Host: example.com\r\n\r\n')
+    with mock.patch.object(
+        web.RequestHandler, "handle_request", create=True, new=handle_request
+    ):
+        srv.data_received(
+            b"GET / HTTP/1.0\r\n" b"Content-Length: 10\r\n" b"Host: example.com\r\n\r\n"
+        )
 
-    await asyncio.gather(srv._task_handler, cancel())
-    assert log.debug.called
+        await asyncio.gather(srv._task_handler, cancel())
+        assert log.debug.called
 
 
 async def test_handle_cancelled(make_srv, transport) -> None:
@@ -525,83 +462,56 @@ async def test_handle_cancelled(make_srv, transport) -> None:
     srv = make_srv(logger=log, debug=True)
     srv.connection_made(transport)
 
-    srv.handle_request = mock.Mock()
     # start request_handler task
     await asyncio.sleep(0)
 
-    srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n\r\n')
+    srv.data_received(b"GET / HTTP/1.0\r\n" b"Host: example.com\r\n\r\n")
 
     r_handler = srv._task_handler
     assert (await r_handler) is None
 
 
 async def test_handle_400(srv, buf, transport) -> None:
-    srv.data_received(b'GET / HT/asd\r\n\r\n')
+    srv.data_received(b"GET / HT/asd\r\n\r\n")
 
     await asyncio.sleep(0)
-    assert b'400 Bad Request' in buf
+    assert b"400 Bad Request" in buf
 
 
-async def test_handle_500(srv, buf, transport, request_handler) -> None:
-    request_handler.side_effect = ValueError
-
-    srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n\r\n')
-    await srv._task_handler
-
-    assert b'500 Internal Server Error' in buf
-
-
-async def test_handle_504(srv, buf, request_handler) -> None:
-    request_handler.side_effect = asyncio.TimeoutError
-
-    srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n\r\n')
-    await srv._task_handler
-
-    assert b'504 Gateway Timeout' in buf
-
-
-async def test_keep_alive(make_srv, transport, ceil) -> None:
+async def test_keep_alive(make_srv, transport) -> None:
     loop = asyncio.get_event_loop()
     srv = make_srv(keepalive_timeout=0.05)
-    srv.KEEPALIVE_RESCHEDULE_DELAY = 0.1
-    srv.connection_made(transport)
+    future = loop.create_future()
+    future.set_result(1)
 
-    srv.keep_alive(True)
-    srv.handle_request = mock.Mock()
-    srv.handle_request.return_value = loop.create_future()
-    srv.handle_request.return_value.set_result(1)
+    with mock.patch.object(
+        web.RequestHandler, "KEEPALIVE_RESCHEDULE_DELAY", new=0.1
+    ), mock.patch.object(
+        web.RequestHandler, "handle_request", create=True, return_value=future
+    ):
+        srv.connection_made(transport)
+        srv.keep_alive(True)
+        srv.data_received(
+            b"GET / HTTP/1.1\r\n" b"Host: example.com\r\n" b"Content-Length: 0\r\n\r\n"
+        )
 
-    srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
+        waiter = None
+        while waiter is None:
+            await asyncio.sleep(0)
+            waiter = srv._waiter
+        assert srv._keepalive_handle is not None
+        assert not transport.close.called
 
-    waiter = None
-    while waiter is None:
-        await asyncio.sleep(0)
-        waiter = srv._waiter
-    assert srv._keepalive_handle is not None
-    assert not transport.close.called
-
-    await asyncio.sleep(0.2)
-    assert transport.close.called
-    assert waiter.cancelled
+        await asyncio.sleep(0.2)
+        assert transport.close.called
+        assert waiter.cancelled
 
 
-async def test_srv_process_request_without_timeout(make_srv,
-                                                   transport) -> None:
+async def test_srv_process_request_without_timeout(make_srv, transport) -> None:
     srv = make_srv()
     srv.connection_made(transport)
 
-    srv.data_received(
-        b'GET / HTTP/1.0\r\n'
-        b'Host: example.com\r\n\r\n')
+    srv.data_received(b"GET / HTTP/1.0\r\n" b"Host: example.com\r\n\r\n")
 
     await srv._task_handler
     assert transport.close.called
@@ -616,25 +526,21 @@ def test_keep_alive_timeout_nondefault(make_srv) -> None:
     assert 10 == srv.keepalive_timeout
 
 
-async def test_supports_connect_method(srv,
-                                       transport, request_handler) -> None:
+async def test_supports_connect_method(srv, transport, request_handler) -> None:
     srv.data_received(
-        b'CONNECT aiohttp.readthedocs.org:80 HTTP/1.0\r\n'
-        b'Content-Length: 0\r\n\r\n')
+        b"CONNECT aiohttp.readthedocs.org:80 HTTP/1.0\r\n" b"Content-Length: 0\r\n\r\n"
+    )
     await asyncio.sleep(0.1)
 
     assert request_handler.called
-    assert isinstance(
-        request_handler.call_args[0][0].content,
-        streams.StreamReader)
+    assert isinstance(request_handler.call_args[0][0].content, streams.StreamReader)
 
 
 async def test_content_length_0(srv, request_handler) -> None:
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.org\r\n'
-        b'Content-Length: 0\r\n\r\n')
-    await asyncio.sleep(0)
+        b"GET / HTTP/1.1\r\n" b"Host: example.org\r\n" b"Content-Length: 0\r\n\r\n"
+    )
+    await asyncio.sleep(0.01)
 
     assert request_handler.called
     assert request_handler.call_args[0][0].content == streams.EMPTY_PAYLOAD
@@ -663,39 +569,7 @@ def test_rudimentary_transport(srv) -> None:
     assert not srv._reading_paused
 
 
-async def test_close(srv, transport) -> None:
-    transport.close.side_effect = partial(srv.connection_lost, None)
-    srv.connection_made(transport)
-    await asyncio.sleep(0)
-
-    srv.handle_request = mock.Mock()
-    srv.handle_request.side_effect = helpers.noop
-
-    assert transport is srv.transport
-
-    srv._keepalive = True
-    srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n'
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
-
-    await asyncio.sleep(0.05)
-    assert srv._task_handler
-    assert srv._waiter
-
-    srv.close()
-    await asyncio.sleep(0)
-    assert srv._task_handler is None
-    assert srv.transport is None
-    assert transport.close.called
-
-
-async def test_pipeline_multiple_messages(
-    srv, transport, request_handler
-):
+async def test_pipeline_multiple_messages(srv, transport, request_handler):
     transport.close.side_effect = partial(srv.connection_lost, None)
 
     processed = 0
@@ -711,12 +585,13 @@ async def test_pipeline_multiple_messages(
 
     srv._keepalive = True
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n'
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
+        b"GET / HTTP/1.1\r\n"
+        b"Host: example.com\r\n"
+        b"Content-Length: 0\r\n\r\n"
+        b"GET / HTTP/1.1\r\n"
+        b"Host: example.com\r\n"
+        b"Content-Length: 0\r\n\r\n"
+    )
 
     assert srv._task_handler is not None
     assert len(srv._messages) == 2
@@ -728,9 +603,7 @@ async def test_pipeline_multiple_messages(
     assert processed == 2
 
 
-async def test_pipeline_response_order(
-    srv, buf, transport, request_handler
-):
+async def test_pipeline_response_order(srv, buf, transport, request_handler):
     transport.close.side_effect = partial(srv.connection_lost, None)
     srv._keepalive = True
 
@@ -741,17 +614,16 @@ async def test_pipeline_response_order(
         await asyncio.sleep(0.01)
         resp = web.StreamResponse()
         await resp.prepare(request)
-        await resp.write(b'test1')
+        await resp.write(b"test1")
         await resp.write_eof()
         processed.append(1)
         return resp
 
     request_handler.side_effect = handle1
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
-    await asyncio.sleep(0)
+        b"GET / HTTP/1.1\r\n" b"Host: example.com\r\n" b"Content-Length: 0\r\n\r\n"
+    )
+    await asyncio.sleep(0.01)
 
     # second
 
@@ -759,17 +631,16 @@ async def test_pipeline_response_order(
         nonlocal processed
         resp = web.StreamResponse()
         await resp.prepare(request)
-        await resp.write(b'test2')
+        await resp.write(b"test2")
         await resp.write_eof()
         processed.append(2)
         return resp
 
     request_handler.side_effect = handle2
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
-    await asyncio.sleep(0)
+        b"GET / HTTP/1.1\r\n" b"Host: example.com\r\n" b"Content-Length: 0\r\n\r\n"
+    )
+    await asyncio.sleep(0.01)
 
     assert srv._task_handler is not None
 
@@ -780,9 +651,8 @@ async def test_pipeline_response_order(
 def test_data_received_close(srv) -> None:
     srv.close()
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
+        b"GET / HTTP/1.1\r\n" b"Host: example.com\r\n" b"Content-Length: 0\r\n\r\n"
+    )
 
     assert not srv._messages
 
@@ -790,9 +660,8 @@ def test_data_received_close(srv) -> None:
 def test_data_received_force_close(srv) -> None:
     srv.force_close()
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: example.com\r\n'
-        b'Content-Length: 0\r\n\r\n')
+        b"GET / HTTP/1.1\r\n" b"Host: example.com\r\n" b"Content-Length: 0\r\n\r\n"
+    )
 
     assert not srv._messages
 
@@ -825,10 +694,7 @@ async def test__process_keepalive_schedule_next(srv) -> None:
     with mock.patch.object(loop, "time", return_value=expire_time):
         with mock.patch.object(loop, "call_later") as call_later_patched:
             srv._process_keepalive()
-            call_later_patched.assert_called_with(
-                1,
-                srv._process_keepalive
-            )
+            call_later_patched.assert_called_with(1, srv._process_keepalive)
 
 
 async def test__process_keepalive_force_close(srv) -> None:
@@ -845,16 +711,48 @@ async def test_two_data_received_without_waking_up_start_task(srv) -> None:
     assert srv._waiter is not None
 
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: ex.com\r\n'
-        b'Content-Length: 1\r\n\r\n'
-        b'a')
+        b"GET / HTTP/1.1\r\n" b"Host: ex.com\r\n" b"Content-Length: 1\r\n\r\n" b"a"
+    )
     srv.data_received(
-        b'GET / HTTP/1.1\r\n'
-        b'Host: ex.com\r\n'
-        b'Content-Length: 1\r\n\r\n'
-        b'b')
+        b"GET / HTTP/1.1\r\n" b"Host: ex.com\r\n" b"Content-Length: 1\r\n\r\n" b"b"
+    )
 
     assert len(srv._messages) == 2
     assert srv._waiter.done()
     await asyncio.sleep(0.01)
+
+
+async def test_client_disconnect(aiohttp_server) -> None:
+    async def handler(request):
+        buf = b""
+        with pytest.raises(ConnectionError):
+            while len(buf) < 10:
+                buf += await request.content.read(10)
+        # return with closed transport means premature client disconnection
+        return web.Response()
+
+    logger = mock.Mock()
+    app = web.Application()
+    app._debug = True
+    app.router.add_route("POST", "/", handler)
+    server = await aiohttp_server(app, logger=logger)
+
+    _, writer = await asyncio.open_connection("127.0.0.1", server.port)
+    writer.write(
+        """POST / HTTP/1.1\r
+Connection: keep-alive\r
+Content-Length: 10\r
+Host: localhost:{port}\r
+\r
+""".format(
+            port=server.port
+        ).encode(
+            "ascii"
+        )
+    )
+    await writer.drain()
+    await asyncio.sleep(0.1)
+    writer.write(b"x")
+    writer.close()
+    await asyncio.sleep(0.1)
+    logger.debug.assert_called_with("Ignored premature client disconnection")
